import json
from operator import itemgetter
import click
from click import secho
from indico_install.utils import run_cmd

from tabulate import tabulate

unit_table = ["k", "m", "g"]


def convert_size(size):
    if size.isdigit():
        size = int(size)
    elif size[:-1].isdigit():
        unit = size[-1].lower()
        size = int(size[:-1]) * (1000 ** (1 + unit_table.index(unit)))
    else:
        raise ValueError()
    for x in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024.0:
            return "%3.1f %s" % (size, x)
        size /= 1024.0

    return size


@click.group("inspect")
@click.pass_context
def inspect(ctx):
    """
    Inspect kubernetes resources
    """
    ctx.ensure_object(dict)


@inspect.command("hpa")
@click.pass_context
def hpa(ctx):
    """
    Outputs a formatted version of Horizontal Pod Autoscalers resource list
    that includes additional information that is usually hidden away in the metadata.
    """
    result = run_cmd(
        """kubectl get hpa -o custom-columns=NAME:.metadata.name,CURRENT:.status.currentReplicas,DESIRED:.status.desiredReplicas,LAST_SCALED:.status.lastScaleTime,MIN:.spec.minReplicas,MAX:spec.maxReplicas,ANNOTATION:.metadata.annotations."autoscaling\.alpha\.kubernetes\.io/current-metrics",TARGETMEM:.metadata.annotations."autoscaling\.alpha\.kubernetes\.io/metrics",CPUTARGET:.spec.targetCPUUtilizationPercentage""",
        silent=True,
    )

    lines = iter(result.split("\n"))
    headers = next(lines).split()
    output_values = []

    for line in lines:
        values = dict(zip(headers, line.split()))
        targets = {}

        # Parse Target Memory
        target_mem = {}
        try:
            target_mem = json.loads(values.pop("TARGETMEM"))
            for target_resource in target_mem:
                if target_resource["type"] != "Resource":
                    continue
                targets[target_resource["resource"]["name"]] = target_resource[
                    "resource"
                ]["targetAverageUtilization"]
        except Exception:
            target_mem["memory"] = target_mem.get("memory", "??")

        targets["cpu"] = values.pop("CPUTARGET")

        try:
            annotation = values.pop("ANNOTATION")
            annotation_json = json.loads(annotation)
            for resource_info in annotation_json:
                try:
                    if resource_info["type"] != "Resource":
                        continue
                    info = resource_info["resource"]
                    name, usage_percent, total_usage = itemgetter(
                        "name", "currentAverageUtilization", "currentAverageValue"
                    )(info)
                    values[
                        f"% {name.upper()} USED"
                    ] = f"{usage_percent}%/{targets[name]}%"
                    values[f"{name.upper()} USAGE"] = total_usage
                except Exception as e:
                    secho(f"Error occured while fetching data: {e.args[0]}")
        except Exception:
            pass

        if "MEMORY USAGE" in values:
            try:
                values["MEMORY USAGE"] = convert_size(values["MEMORY USAGE"])
            except Exception:
                pass
        output_values.append(values)

    headers = [
        "NAME",
        "CURRENT",
        "DESIRED",
        "MIN",
        "MAX",
        "% CPU USED",
        "CPU USAGE",
        "% MEMORY USED",
        "MEMORY USAGE",
    ]
    output_lines = []
    for values in output_values:
        output_line = []
        for item in headers:
            output_line.append(values.pop(item, "<N/A>"))
        output_lines.append(output_line)
    secho(tabulate(output_lines, headers=headers))



import click

from indico_install.helm.render import render


@click.command("put-resource")
@click.pass_context
@click.argument("service", required=True, type=str)
@click.option(
    "-l",
    "--limit",
    required=True,
    type=str,
    help="An int/float with '[E/P/T/G/M/K]i' mem or (optional) 'm' cpu suffix",
)
@click.option(
    "-r",
    "--resource",
    default="memory",
    type=click.Choice(["memory", "cpu"]),
    show_default=True,
    help="Resource to update limits for",
)
def put_resource(ctx, service, *, limit, resource="memory"):
    """
    Provide a new resource limit for limit for memory or cpu

    ARGS:

        <SERVICE> exact name of service to update
    """
    valid_limit = limit
    err = lambda e: click.secho(f"Invalid {resource} limit: {limit}. {e}", fg="red")
    if resource == "memory":
        if not limit[-2:] in ["Ei", "Pi", "Ti", "Gi", "Mi", "Ki"]:
            err("Mem limit must end in power-of-two suffix (Gi, Mi, etc)")
            return
        valid_limit = valid_limit[:-2]
    else:
        _suff = limit[-1:]
        if not (_suff.isdigit() or _suff == "m"):
            err("CPU limit must end in digit or 'm' suffix")
            return
        elif _suff == "m":
            valid_limit = valid_limit[:-1]

    try:
        float(valid_limit)
    except Exception:
        err("Limit must be a valid int or float")
        return

    ctx.obj["TRACKER"].edit_cluster_config(
        changes={"services": {service: {"values": {"resources": {resource: limit}}}}}
    )
    ctx.obj["TRACKER"].save()
    ctx.invoke(
        render,
        cluster_manager=ctx.obj["TRACKER"],
        services=[service],
        allow_image_overrides=True,
    )
    click.secho(
        f"Rendered template for {service}. Don't forget to apply!", fg="green",
    )



import click

from indico_install.helm.render import render
from indico_install.utils import find_names, run_cmd


@click.command("enable")
@click.pass_context
@click.argument("service", required=True, type=str)
def enable(ctx, service):
    """
    Enable a K8S deployment/statefulset/service.
    Will also render the templates - you must apply on your own

    ARGS:
        <SERVICE> exact name of service to enable
    """
    click.secho(
        f"Setting {service}.disabled=False in cluster config", fg="blue",
    )
    ctx.obj["TRACKER"].edit_cluster_config(
        changes={"services": {service: {"<!disabled>": False}}}
    )
    ctx.obj["TRACKER"].save()
    ctx.invoke(
        render,
        cluster_manager=ctx.obj["TRACKER"],
        services=[service],
    )
    click.secho(
        f"Rendered template for {service}. Don't forget to apply!", fg="green",
    )


@click.command("delete")
@click.pass_context
@click.argument("service", required=True, type=str)
def delete(ctx, service):
    """
    Delete a K8S deployment/statefulset/service and any associated pdbs/hpas

    ARGS:
        <SERVICE> grep string of deployments, statefulsets, services to delete
    """
    updated_svcs = []
    for svc_type in ["deployment", "statefulset", "service", "pdb", "hpa"]:
        out = find_names(svc_type, service)
        if not out:
            continue
        for _svc in out:
            if click.confirm(f"Ok to delete {svc_type} {_svc}"):
                click.secho(run_cmd(f"kubectl delete {svc_type} {_svc}"), fg="green")
                if svc_type not in ("hpa", "pdb"):
                    updated_svcs.append(_svc)

    if updated_svcs:
        ctx.obj["TRACKER"].edit_cluster_config(
            changes={"services": {_svc: {"<!disabled>": True} for _svc in updated_svcs}}
        )
        ctx.obj["TRACKER"].save()


@click.command("scale")
@click.pass_context
@click.argument("service", required=True, type=str)
@click.argument("amount", required=True, type=int)
@click.option(
    "--contains/--no-contains",
    default=True,
    show_default=True,
    help=("Scale all services with names containing the <service> string."),
)
def scale(ctx, service, amount, contains=True):
    """
    Scale a K8S cluster deployment or statefulset

    ARGS:
        <SERVICE> grep string of deployments and statefulsets to scale

        <AMOUNT> number of pods to create
    """
    updated_svcs = []
    for svc_type in ["deployment", "statefulset"]:
        out = find_names(svc_type, service, contains)
        if not out:
            continue
        for _svc in out:
            click.secho(
                run_cmd(f"kubectl scale --replicas={amount} {svc_type} {_svc}"),
                fg="green",
            )
            hpa_exists = run_cmd(
                f"kubectl get hpa {_svc} -o json 2>/dev/null", silent=True
            )
            if hpa_exists:
                click.secho(
                    run_cmd(
                        "kubectl patch hpa %s --patch "
                        """'{"spec":{"minReplicas": %s, "maxReplicas": %s}}'"""
                        % (_svc, amount, amount + 2)
                    ),
                    fg="green",
                )
            updated_svcs.append(_svc)

    if updated_svcs:
        ctx.obj["TRACKER"].edit_cluster_config(
            changes={
                "services": {
                    _svc: {"values": {"replicas": amount}} for _svc in updated_svcs
                }
            }
        )
        ctx.obj["TRACKER"].save()



import shutil
import tarfile
import time
from pathlib import Path

import click
from click import secho

from indico_install.config import CLUSTER_NAME, yaml, D_PATH
from indico_install.infra.input_utils import auth_with_gsutil
from indico_install.utils import current_user, run_cmd
from indico_install.cluster_manager import ClusterManager, get_default_namespace


default_namespace = get_default_namespace(ClusterManager().cluster_config, "defaultNamespace")


@click.command("cluster-info")
@click.pass_context
@click.option(
    "--upload/--no-upload",
    default=True,
    show_default=True,
    help="Upload the cluster dump TAR to a private Google Storage bucket shared with Indico",
)
@click.option(
    "--sudo/--no-sudo",
    default=False,
    show_default=False,
    help="Use sudo to run cluster dump (useful if running into permission issues)",
)
def cluster_info(ctx, *, upload, sudo):
    """
    Load and package all the information on the running cluster
    into a local TAR file.
    """
    dumpdir = D_PATH / f"cluster_info_{time.strftime('%Y%m%d-%H%M%S')}"
    secho("Pulling cluster info...", fg="blue")
    sudo = "sudo " if sudo else ""
    run_cmd(
        f"""
        {sudo}kubectl cluster-info dump -o yaml --output-directory={dumpdir};
        {sudo}kubectl get configmaps --namespace={default_namespace} -o yaml > {dumpdir}/configmaps.yaml;
        {sudo}kubectl get ingress --namespace={default_namespace} -o yaml > {dumpdir}/ingress.yaml;
        {sudo}kubectl get pv --namespace={default_namespace} -o yaml > {dumpdir}/pv.yaml;
        {sudo}kubectl get pvc --namespace={default_namespace} -o yaml > {dumpdir}/pvc.yaml;
        """,
        silent=True,
    )
    # Remove cluster mananger from CMs
    with open(f"{dumpdir}/configmaps.yaml", "r") as f:
        _configs = yaml.safe_load(f)
    with open(f"{dumpdir}/configmaps.yaml", "w") as f:
        try:
            cm_index = next(
                i
                for i, cm in enumerate(_configs["items"])
                if cm["metadata"]["name"].startswith("cluster-manager")
            )
            _configs["items"].pop(cm_index)
            yaml.dump(_configs, f, default_flow_style=False)
        except Exception:
            pass

    dump_tar = str(dumpdir.resolve()) + ".tar.gz"
    with tarfile.open(dump_tar, "w:gz") as tar:
        tar.add(dumpdir, arcname="")

    secho(f"Cluster information saved to {dump_tar}", fg="green")
    shutil.rmtree(dumpdir, ignore_errors=True)
    if not upload:
        return

    cluster_user = (
        current_user(clean=True) if auth_with_gsutil() else CLUSTER_NAME
    )
    if not cluster_user:
        secho(
            "Cannot authenticate with Google - please upload cluster info manually",
            fg="yellow",
        )
        return

    result = run_cmd(
        f"gsutil cp {dump_tar} gs://client_{cluster_user}/{Path(dump_tar).name} 2>&1",
        silent=True,
    )
    msgs = (
        (
            "Operation completed",
            "Cluster information successfully shared with Indico",
            "green",
        ),
        (
            "BucketNotFoundException",
            f"Bucket for user {cluster_user} does not exist. Please ask Indico to create this and try sharing again, or manually share your cluster info TAR",
            "yellow",
        ),
        (
            "",
            "Unable to share cluster info with Indico. Please share the cluster info TAR manually",
            "red",
        ),
    )
    for res, msg, color in msgs:
        if res in result:
            secho(msg, fg=color)
            break



environments = {
    "production": {
        "abbrev": "prod",
        "color": "red",
        "network": {
            "name": "indico-production",
            "us-central1-f": "production-central1",
            "us-east4-c": "production-east4",
        },
    },
    "development": {
        "abbrev": "dev",
        "color": "green",
        "network": {
            "name": "indico-development",
            "us-central1-f": "development-central1",
            "us-east4-c": "development-default",
        },
    },
}
environments.update(
    {
        "staging": {
            "abbrev": "stage",
            "color": "blue",
            "network": environments["production"]["network"],
        },
        "feature": {
            "abbrev": "feat",
            "color": "yellow",
            "network": environments["development"]["network"],
        },
    }
)



import json
import time

import click

from indico_install.cluster_manager import ClusterManager
from indico_install.config import CLUSTER_NAME
from indico_install.utils import get_cluster_nodes, run_cmd


@click.group("cluster")
@click.pass_context
def cluster(ctx):
    """
    Manage up and down state of cluster
    """
    pass


def get_node_pools_for_scaling(cluster_name, format=None):
    cluster_found = run_cmd(
        "gcloud container clusters list"
        " --filter=resourceLabels.ephemeral=true"
        f" --filter name={cluster_name}"
        " 2> /dev/null"
        " | tail -n 1 | awk '{print $1}'",
        silent=True,
    )

    if cluster_found != cluster_name:
        raise click.UsageError(
            f"Ephemeral Cluster {cluster_name} not found. Found {cluster_found}"
        )

    if not format:
        return run_cmd(
            f"gcloud container node-pools list"
            f" --cluster={cluster_name}"
            ' --quiet --format="value(name)"',
            silent=True,
        ).splitlines()
    return run_cmd(
        f"gcloud container node-pools list"
        f" --cluster={cluster_name}"
        f' --quiet --format="{format}"',
        silent=True,
    )


def wait_status(cluster_name):
    while True:
        status = run_cmd(
            f"gcloud container clusters describe {cluster_name}"
            f" --format 'value(status)'"
            f" --quiet",
            silent=True,
        ).lower()

        if status == "running":
            return
        elif status == "reconciling":
            pass
        else:
            click.secho(f"Unrecognized status: {status}", fg="yellow")

        time.sleep(5)


@cluster.command("up")
@click.option(
    "-c",
    "--cluster-name",
    default=CLUSTER_NAME,
    show_default=True,
    help="Name of cluster",
)
@click.pass_context
def up(ctx, cluster_name):
    """
    Turn scale existing node pools back up from down state by configuration or defaults
    Example Config:
    nodePools:
        default-gpu-pool:
            size: 2
            autoscale: false
        default-gpu-pvm-pool:
            size: 0
            autoscale:
                min: 0
                max: 6
        default-pool:
            size: 2
            autoscale: false
        default-pvm-pool:
            size: 0
            autoscale:
                min: 0
                max: 6
    """
    config = ClusterManager().to_dict()
    node_pool_configs = config["cluster_config"].get("nodePools", {})
    node_pools = get_node_pools_for_scaling(cluster_name)
    # organized so non-autoscaling pools come up first
    pools = sorted(
        [
            (
                pool,
                node_pool_configs.get(pool)
                or {
                    "size": 0 if "pvm" in pool else 2,
                    "autoscale": {"min": 0, "max": 6} if "pvm" in pool else False,
                },
            )
            for pool in node_pools
        ],
        key=lambda x: bool(x[1]["autoscale"]),
    )

    for pool, pool_config in pools:
        wait_status(cluster_name)

        run_cmd(
            f"gcloud container clusters resize {cluster_name}"
            f" --node-pool={pool}"
            f" --num-nodes={pool_config['size']}"
            " --quiet"
        )

        autoscale_config = pool_config["autoscale"]
        if autoscale_config:
            max_size = autoscale_config["max"]
            min_size = autoscale_config["min"]
            wait_status(cluster_name)
            run_cmd(
                f"gcloud container clusters update {cluster_name}"
                f" --node-pool={pool}"
                " --enable-autoscaling"
                f" --max-nodes {max_size}"
                f" --min-nodes {min_size}"
                " --quiet"
            )


@cluster.command("down")
@click.option(
    "-c",
    "--cluster-name",
    default=CLUSTER_NAME,
    show_default=True,
    help="Name of cluster",
)
@click.pass_context
def down(ctx, cluster_name):
    """
    Turn scale existing node pools down from up state
    """
    node_pools = json.loads(get_node_pools_for_scaling(cluster_name, format="json"))
    curr_set = {}

    # 0. Track current pool settings
    for pool in node_pools:
        if pool.get("autoscaling", {}).get("enabled"):
            #since we are no longer using pvm nodes, we only use autoscaling groups. These need to start with an initial count.
            size = pool["initialNodeCount"]
            autoscale = {"min": 0, "max": pool["autoscaling"]["maxNodeCount"]}
        else:
            size = pool["initialNodeCount"]
            autoscale = False
        
        curr_set[pool['name']] = {"size": size, "autoscale": autoscale}

    # 1. Record current state to CM if possible
    cm = ClusterManager()
    if cm.cm_exists:
        cm.edit_cluster_config(changes={"nodePools": curr_set})
        cm.save()

    # 2. Disable any autoscaling pools
    for pool in curr_set:
        if curr_set[pool]["autoscale"]:
            wait_status(cluster_name)
            run_cmd(
                f"gcloud container clusters update {cluster_name}"
                f" --node-pool={pool}"
                " --no-enable-autoscaling"
                " --quiet"
            )

    # 3. Cordon and drain all nodes
    nodes = get_cluster_nodes()
    for node in nodes:
        run_cmd(f"kubectl cordon {node}")
    for node in nodes:
        run_cmd(
            "kubectl drain --delete-local-data --ignore-daemonsets "
            f"--disable-eviction {node}",
            silent=True,
        )

    # 4. Resize node pools
    for pool in node_pools:
        wait_status(cluster_name)
        run_cmd(
            f"gcloud container clusters resize {cluster_name}"
            f" --node-pool={pool['name']}"
            " --num-nodes=0"
            " --quiet"
        )



import os
import shutil
import tempfile

import click
from indico_install.cluster_manager import ClusterManager
from indico_install.config import D_PATH, REMOTE_TEMPLATES_PATH, yaml
from indico_install.utils import options_wrapper, run_cmd
from indico_install.utils import run_cmd


def helm3_install(dependency, default_values_path):
    """
    Helm3 install a given chart
    """
    click.echo(f"Installing {dependency['name']}")
    if not all(key in dependency.keys() for key in ["name", "repository", "version"]):
        click.secho(
            f"Unable to install {dependency.get('name')}: expected keys not present"
        )
        return
    default_values = (
        "-f " + str(default_values_path) if default_values_path.is_file() else ""
    )
    repo_name = dependency.get("repoName", f"{dependency['name']}-repository")
    namespace = (
        f"--namespace {dependency.get('namespace')}"
        if dependency.get("namespace")
        else ""
    )
    command = "upgrade" if release_exists(dependency) else "install"
    wait = "--wait" if dependency.get("wait", True) else ""
    args = dependency.get("args", "")
    with tempfile.NamedTemporaryFile(suffix=".tmp", delete=False) as override_yaml:
        override_yaml.write(
            yaml.dump(dependency.get("values", {}), default_flow_style=False).encode(
                "utf-8"
            )
        )
        override_yaml.flush()
        run_cmd(f"helm3 repo add {repo_name} {dependency['repository']}")
        run_cmd("helm3 repo update", silent=True)
        run_cmd(
            f"helm3 {command} {namespace} {wait} {args} {dependency['name']} {repo_name}/{dependency['name']} --create-namespace {default_values} -f {override_yaml.name} --version {dependency['version']}"
        )


def release_exists(dependency):
    """
    Given a dependency, return true if it is deployed, else return false
    """
    output = run_cmd(
        f"helm3 get all {dependency.get('name')} -n {dependency.get('namespace', 'default')}",
        silent=True,
    )
    if len(output) > 0:
        click.secho(
            f"Release {dependency.get('name')} already exists; updating existing release"
        )
        return True
    return False


def fetch_defaults(remote_path):
    """
    Fetch default helm values from updraft
    """
    remote_helm_values_path = (
        REMOTE_TEMPLATES_PATH + remote_path + "/helm-values.tar.gz"
    )
    local_directory = (
        D_PATH / "helm-values" / "".join(c for c in str(remote_path) if c.isalnum)
    )
    if not local_directory.parent.exists():
        local_directory.parent.mkdir(exist_ok=True, parents=True)
    if local_directory.is_dir():
        shutil.rmtree(local_directory)
    click.secho(
        f"Downloading indico default helm values to {local_directory}", fg="yellow"
    )
    os.makedirs(local_directory, exist_ok=True)
    run_cmd(
        f"wget {remote_helm_values_path} -O - | " f"tar -xz -C {local_directory}",
        silent=False,
    )
    assert (local_directory).is_dir(), f"Unable to download version {remote_path}"
    return local_directory / "helm-values"


@click.command("install")
@click.argument("charts", required=False, nargs=-1)
@click.pass_context
@options_wrapper()
def install(ctx, cluster_manager=None, charts=None, yes=False):
    """
    Install helm charts from sources defined in the 'dependencies' section of the cluster manager config. Install any number of available charts with the CHART argument

    Example 1
    Install the cert-manager helm chart as specified in the cluster manager config:
    indico install cert-manager

    Example 2
    Install all helm charts specified in the cluster manager config:
    indico install
    """
    cluster_manager = cluster_manager or ClusterManager()
    helm_defaults_path = fetch_defaults(cluster_manager.indico_version)
    deps = cluster_manager.cluster_config.get("dependencies", {})
    if not deps:
        click.secho(
            "No helm chart dependencies specified in the cluster manager configmap"
        )
        return
    for dependency in deps:
        if charts and not [
            chart for chart in charts if chart in dependency.get("name")
        ]:
            continue
        if yes or click.confirm(
            f"Ready to install/upgrade {dependency.get('name')} chart?"
        ):
            helm3_install(dependency, helm_defaults_path)



#!/usr/bin/env python3
import os
from subprocess import check_call

from setuptools import find_packages, setup

setup(
    name="indico_install",
    version="2.1.6",
    description="An indico install platform",
    author="Indico",
    author_email="engineering@indico.io",
    packages=find_packages(),
    install_requires=[
        "click>=8.0.1",
        "pyyaml>=5.0.0",
        "requests>=2.25.0",
        "cached_property>=1.5.2",
        "google-cloud-storage>=1.25.0",
        "tabulate>=0.8.6",
        "untangle>=1.1.1",
        "azure-cli>=2.28.0",
        "boto3>=1.18.40",
        "whatthepatch>=1.0.2",
        "python3-saml>=1.12.0"
    ],
    extras_require={
        "aks": ["azure-cli>=2.25.0"],
        "eks": ["boto3>=1.9.176","awscli>=1.18.152"],
        "indico": ["whatthepatch>=1.0.2"],
        "saml": ["python3-saml>=1.9.0"],
    },
    include_package_data=True,
    entry_points="""
        [console_scripts]
        indico=indico_install:indico_cli
    """,
    scripts=["bin/kube", "bin/cluster_info.sh"]
            + [f"bin/kube_bin/{filename}" for filename in os.listdir("bin/kube_bin")],
)

check_call(
    [
        "bash",
        "-c",
        'azcomm=$(command -v az); if ! [ -z "$azcomm" ]; then sed -i "s/^python /python3 /g" $azcomm; fi',
    ]
)
